/*==============================================================================*/
/*  Copyright (C) 2009-2016, Heartland Data inc. All Rights Reserved.           */
/*                                                                              */
/*  Title  :   SDIF Driver                                                      */
/*  EventID:   Insert Kernelinfo Test Point                                     */
/*  AppVer :   DT10 Ver11.XX`@                                                */
/*  FileID :   41a1                                                             */
/*  Version:   1.1                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/vmalloc.h>
#include <linux/version.h>

#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/core.h>

#define	PROC_NAME	"dt10sdif"

#define MAX_SD_LOG_SIZE 32*1024*1024
//#define REAL_TIME_KERNEL_THREAD
//#define DT10_USE_RING_BUFFER

#define DT_ADD_TIME_INFO	1

struct DT10_DATA {
	char tp_func;
	unsigned int addr;
	unsigned int dat;
	unsigned int size;
	unsigned char value[256];
};

#define TP_BusOut		'3'
#define TP_MemoryOutput		'4'

static struct mmc_host *dt10sd_host = NULL;
static struct task_struct *dt10kthread = NULL;
static u32 start_time;
static int dt10busout_flag = 0;
static unsigned char *sdlog = NULL;
static int logWritePos = 0;
static int logWriteLen = 0;
static int logReadPos = 0;
static int logReadLen = 0;
#ifdef DT10_USE_RING_BUFFER
static int logWriteNum = 0;
#endif
static int printed = 0;
static DEFINE_SPINLOCK(tplock);
static int tid = 0;

void dt10_sd_host_get(struct mmc_host *host)
{
	dt10sd_host = host;
}
EXPORT_SYMBOL(dt10_sd_host_get);

static inline void put_sdlog(u8 dat)
{
	*(sdlog + logWritePos) = dat;
	++logWritePos;
	++logWriteLen;
#ifdef DT10_USE_RING_BUFFER
	if (logWritePos >= MAX_SD_LOG_SIZE) {
		logWritePos = 0;
		++logWriteNum;
	}
#endif
}

static inline u8 get_sdlog(void)
{
	u8 dat;

	dat = *(sdlog + logReadPos);
	++logReadPos;
	++logReadLen;
#ifdef DT10_USE_RING_BUFFER
	if (logReadPos >= MAX_SD_LOG_SIZE) {
		logReadPos = 0;
		--logWriteNum;
	}
#endif
	return dat;
}

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
void _TP_BusOut(unsigned int addr, unsigned int data)
{
	unsigned long flags;
	u8 tmp;
#if DT_ADD_TIME_INFO
	struct timespec now;
	u32 time;
#endif
	if (dt10busout_flag != 1) return;
	if (sdlog == NULL) return;
	if (logWriteLen == 0) {
		printk( KERN_INFO "DT10 Trace Start\n");
	}
#ifdef DT10_USE_RING_BUFFER
	if (logWriteNum > 0 && logWritePos >= logReadPos) {
		printk( KERN_INFO "DT10 SDIF Buffer Overflow by Busout\n");
		return;
	}
#else
	if (logWriteLen + 8 >= MAX_SD_LOG_SIZE) {
		if (printed == 0) {
			printk( KERN_INFO "DT10 Trace Stop\n");
			printed = 1;
		}
		return;
	}
#endif	

	spin_lock_irqsave(&tplock, flags);

#if DT_ADD_TIME_INFO
	getnstimeofday(&now);
	time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	time -= start_time;
#endif
	put_sdlog(41);	/* Normal TP opcode */
	tmp = (data >> 8) & 0x3F;
	tmp += smp_processor_id() << 6;	//CPU ID
	put_sdlog(tmp);
	put_sdlog(data);
	/* setting TP mode bit */
	tmp = addr >> 12;
#if DT_ADD_TIME_INFO
	tmp |= 0x40;
#endif
	put_sdlog(tmp);
	put_sdlog(addr >> 4);
#if DT_ADD_TIME_INFO
	put_sdlog(time);
	put_sdlog(time >> 8);
	put_sdlog(time >> 16);
	put_sdlog(time >> 24);
#endif
	spin_unlock_irqrestore(&tplock, flags);
}
EXPORT_SYMBOL(_TP_BusOut);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput(unsigned int addr, unsigned int data, void *value, unsigned int size)
{
	int i;
	u8 *p;
	unsigned long flags;
	u32 tmp;
#if DT_ADD_TIME_INFO
	struct timespec now;
	u32 time;
#endif

	if (dt10busout_flag != 1) return;
	if (sdlog == NULL) return;
	if (logWriteLen == 0) {
		printk( KERN_INFO "DT10 Trace Start\n");
	}
#ifdef DT10_USE_RING_BUFFER
	if (logWriteNum > 0 && logWritePos >= logReadPos) {
		printk( KERN_INFO "DT10 SDIF Buffer Overflow by MemoryOutput\n");
		return;
	}
#else
	if (logWriteLen + (12 + (size)) >= MAX_SD_LOG_SIZE) {
		if (printed == 0) {
			printk( KERN_INFO "DT10 Trace Stop\n");
			printed = 1;
		}
		return;
	}
#endif

	spin_lock_irqsave(&tplock, flags);

#if DT_ADD_TIME_INFO
	getnstimeofday(&now);
	time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	time -= start_time;
#endif
	put_sdlog(51);		/* Variable TP opcode */
	tmp = (data >> 8) & 0x3F;
	tmp += smp_processor_id() << 6;	//CPU ID
	put_sdlog(tmp);
	put_sdlog(data);
	tmp = addr >> 12;
#if DT_ADD_TIME_INFO
	tmp |= 0x40;			// Time bit
#endif
	put_sdlog( tmp );
	put_sdlog( addr >> 4);
#if DT_ADD_TIME_INFO
	put_sdlog(time);
	put_sdlog(time >> 8);
	put_sdlog(time >> 16);
	put_sdlog(time >> 24);
#endif
	if (tmp & 0x80) {
		put_sdlog(tid);
		put_sdlog(tid >> 8);
		put_sdlog(tid >> 16);
		put_sdlog(tid >> 24);
	}

	if( size > 256 ) size = 256;
	put_sdlog( (size+3)/4 );	// Packet Size
	put_sdlog( size );		// Var Size
	p = (u8 *)value;
	for( i = 0; i < size; ++i, ++p ){
		put_sdlog( *p );
	}
	spin_unlock_irqrestore(&tplock, flags);
}
EXPORT_SYMBOL(_TP_MemoryOutput);

/*==============================================================================*/
/*  Func:   _TP_BusKernelInfo                                                   */
/*  Desc:   Called by KernelInfo TestPoint                                      */
/*==============================================================================*/
void _TP_BusKernelInfo(unsigned int addr, unsigned int dat, struct task_struct *next)
{
	unsigned char knlinfo[20];

	if (dt10busout_flag != 1) return;
	tid = next->pid;
	*(pid_t *)&knlinfo[0] = next->mm?next->tgid:0;
	memcpy( &knlinfo[4], next->comm, TASK_COMM_LEN );

	_TP_MemoryOutput((addr | 0x80000), dat, (void *)&knlinfo, sizeof(knlinfo));
}
EXPORT_SYMBOL(_TP_BusKernelInfo);

/*==============================================================================*/
/*  Func:   _TP_EventTrigger32                                                  */
/*  Desc:   Called by Event Trigger 32bit TestPoint                             */
/*==============================================================================*/
void _TP_EventTrigger32(unsigned int addr, unsigned int data, unsigned int event_id)
{
        unsigned long flags;
        u8 tmp;
#if DT_ADD_TIME_INFO
        struct timespec now;
        u32 time;
#endif

        if (dt10busout_flag != 1) return;
        if (sdlog == NULL) return;
        if (logWriteLen == 0) {
                printk( KERN_INFO "DT10 Trace Start\n");
        }
#ifdef DT10_USE_RING_BUFFER
        if (logWriteNum > 0 && logWritePos >= logReadPos) {
                printk( KERN_INFO "DT10 SDIF Buffer Overflow by Busout\n");
                return;
        }
#else
        if (logWriteLen + 8 >= MAX_SD_LOG_SIZE) {
                if (printed == 0) {
                        printk( KERN_INFO "DT10 Trace Stop\n");
                        printed = 1;
                }
                return;
        }
#endif

        spin_lock_irqsave(&tplock, flags);

#if DT_ADD_TIME_INFO
        getnstimeofday(&now);
        time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
        time -= start_time;
#endif
        put_sdlog(41);  /* Normal TP opcode */
        tmp = (data >> 8) & 0x3F;
        tmp += smp_processor_id() << 6; //CPU ID
        put_sdlog(tmp);
        put_sdlog(data);
        /* setting TP mode bit */
        tmp = addr >> 12;
#if DT_ADD_TIME_INFO
        tmp |= 0x40;
#endif
        tmp |= 0x80;
        put_sdlog(tmp);
        put_sdlog(addr >> 4);
#if DT_ADD_TIME_INFO
        put_sdlog(time);
        put_sdlog(time >> 8);
        put_sdlog(time >> 16);
        put_sdlog(time >> 24);
#endif
        put_sdlog(event_id);
        put_sdlog(event_id >> 8);
        put_sdlog(event_id >> 16);
        put_sdlog(event_id >> 24);

        spin_unlock_irqrestore(&tplock, flags);
}
EXPORT_SYMBOL(_TP_EventTrigger32);

static int dt10busout(void *arg)
{
	struct mmc_command cmd;
	int i, j, size, packet, data_count;
	unsigned long flags;
	u8 tmp, opcode;

	while (!kthread_should_stop()) {
#ifdef REAL_TIME_KERNEL_THREAD
		if (dt10busout_flag == 1 && dt10sd_host != NULL) {
#else
		if (dt10busout_flag == 2 && dt10sd_host != NULL) {
#endif
			if (logReadPos >= MAX_SD_LOG_SIZE || logReadPos >= logWritePos) {
				schedule();
				continue;
			}

			if( mmc_try_claim_host(dt10sd_host) == 0 ){
				schedule();
				continue;
			}

			memset(&cmd, 0, sizeof(struct mmc_command));
			cmd.flags = MMC_RSP_NONE;

			/* sending TP Info  */
			spin_lock_irqsave(&tplock, flags);
			opcode = get_sdlog();
			cmd.arg  = get_sdlog() << 24;
			cmd.arg += get_sdlog() << 16;
			tmp = get_sdlog();
			cmd.arg += tmp << 8;
			cmd.arg += get_sdlog();
			spin_unlock_irqrestore(&tplock, flags);
			cmd.opcode = opcode;
			mmc_wait_for_cmd(dt10sd_host, &cmd, 3);

#if DT_ADD_TIME_INFO
			/* sending Time Info  */
			if (tmp & 0x40) {
				spin_lock_irqsave(&tplock, flags);
				cmd.arg  = get_sdlog() << 24;
				cmd.arg += get_sdlog() << 16;
				cmd.arg += get_sdlog() << 8;
				cmd.arg += get_sdlog();
				spin_unlock_irqrestore(&tplock, flags);
				cmd.opcode = 43;
				mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
			}
#endif

			/* Senging Event Info */
			if (tmp & 0x80) {
				spin_lock_irqsave(&tplock, flags);
				cmd.arg  = get_sdlog() << 24;
				cmd.arg += get_sdlog() << 16;
				cmd.arg += get_sdlog() << 8;
				cmd.arg += get_sdlog();
				spin_unlock_irqrestore(&tplock, flags);
				cmd.opcode = 44;
				mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
			}

			/* sending Value Info  */
			if (opcode == 51) {
				spin_lock_irqsave(&tplock, flags);
				packet = get_sdlog();
				size = get_sdlog();
				spin_unlock_irqrestore(&tplock, flags);
				cmd.arg = packet << 8;
				cmd.arg += size;
				cmd.opcode = 45;	/* Var Size */
				mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
				data_count = 0;
				for (i = 0; i < packet; i++) {
					cmd.arg = 0;
					spin_lock_irqsave(&tplock, flags);
					for (j = 0; j < 4; j++) {
						cmd.arg <<= 8;
						if (data_count < size) {
							cmd.arg += get_sdlog();
							data_count++;
						}
					}
					spin_unlock_irqrestore(&tplock, flags);
					cmd.opcode = 46;	/* Var Data */
					mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
				}
			}
			mmc_release_host(dt10sd_host);
		}
	}
	return 0;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
static ssize_t proc_write(struct file *filp, const char __user *buf, size_t len, loff_t *data)
#else
static int proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
#endif
{
	struct timespec now;
	struct DT10_DATA dt10_data;

	if (copy_from_user(&dt10_data, buf, len)) {
		return -EFAULT;
	}

	switch (dt10_data.tp_func) {
		case '1':
			dt10busout_flag = 1;
			getnstimeofday(&now);
			start_time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
			break;
		case 's':
#ifndef REAL_TIME_KERNEL_THREAD
			dt10busout_flag = 2;
			dt10kthread = kthread_run(dt10busout, NULL, "dt10busout" );
			if (IS_ERR(dt10kthread))
				return -1;
   		 		printk(KERN_INFO "pid->%d:prio->%d:comm->%s\n",
    	          	 dt10kthread->pid,
        	      	 dt10kthread->static_prio,
            	  	 dt10kthread->comm);
#endif
			printk( KERN_INFO "logWritePos=%d\n", logWritePos);
			printk( KERN_INFO "logReadPos=%d\n", logReadPos );
			printk( KERN_INFO "logWriteLen=%d\n", logWriteLen);
			printk( KERN_INFO "logReadLen=%d\n", logReadLen );
			break;
		case TP_BusOut:
			_TP_BusOut(dt10_data.addr, dt10_data.dat);
			break;
		case TP_MemoryOutput:
			_TP_MemoryOutput(dt10_data.addr, dt10_data.dat, dt10_data.value, dt10_data.size);
			break;
		default:
#ifndef REAL_TIME_KERNEL_THREAD
			if (dt10kthread) {
				kthread_stop(dt10kthread);
				dt10kthread = NULL;
			}
#endif
			dt10busout_flag = 0;
			logWritePos = 0;
			logWriteLen = 0;
			logReadPos = 0;
			logReadLen = 0;
			printed = 0;
#ifdef DT10_USE_RING_BUFFER
			logWriteNum = 0;
#endif
			break;
	}

	return len;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
static ssize_t proc_read(struct file *filp, char __user *buf, size_t count, loff_t *data)
#else
static int proc_read(char *buf, char **start, off_t offset, int count, int *eof, void *data)
#endif
{
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
	*eof = 1;
#endif
	return count;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
static const struct file_operations dt_proc_fops = {
	.write = proc_write,
	.read = proc_read,
};
#endif

static int __init dt10_sdif_init(void)
{
	struct proc_dir_entry* entry;

	sdlog = vmalloc(MAX_SD_LOG_SIZE);	
	if( sdlog == NULL ){
	    printk(KERN_INFO "DT10 Cannot alloc vmalloc\n");
	}	

#ifdef REAL_TIME_KERNEL_THREAD
	dt10kthread = kthread_run(dt10busout, NULL, "dt10busout" );
	if (IS_ERR(dt10kthread))
		return -1;
    	printk(KERN_INFO "pid->%d:prio->%d:comm->%s\n",
               dt10kthread->pid,
               dt10kthread->static_prio,
               dt10kthread->comm);
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
	entry = proc_create(PROC_NAME, 666, NULL, &dt_proc_fops);
#else
	entry = create_proc_entry(PROC_NAME, 0666, NULL);
#endif
	if ( !entry ) {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
	entry->write_proc = proc_write;
	entry->read_proc = proc_read;
#endif

	printk( KERN_INFO "dt10sdif load OK\n" );
	return 0;
}

static void __exit dt10_sdif_exit(void)
{
#ifdef REAL_TIME_KERNEL_THREAD
	if (dt10kthread) {
		kthread_stop(dt10kthread);
		dt10kthread = NULL;
	}
#endif
	vfree(sdlog);
	remove_proc_entry(PROC_NAME, NULL);
	printk( KERN_INFO "dt10sdif unload\n" );
}

module_init( dt10_sdif_init );
module_exit( dt10_sdif_exit );

MODULE_DESCRIPTION("dt10sdif");
MODULE_LICENSE("GPL2");
